(*  BarnesRoutines.m 

    David A. Kosower

    Version 1.0.0, November 6, 2007

      Package to automate application of Barnes lemmas to Mellin-Barnes
   integrals.  Intended for use with the output of Czakon's MB.

      There are five externally-visible routines,
   DoBarnes1; DoBarnes2; DoAllBarnes; ExpandInt and SetVerbosity.

   The first two apply the first and second Barnes lemmas to a Mellin-Barnes
   integral, represented as in MB by an MBint object.  Both try to make
   linear changes of variables in order to enable the use of the Barnes
   lemmas.  The application of the Barnes lemmas includes some derivatives
   of the first Barnes lemma, and some corollaries of the second due to
   V. A. Smirnov.  The third routine, DoAllBarnes, applies the Barnes lemmas
   repeatedly to a list of functions.  It has an optional second argument,
   which if True, applies the lemmas to integrals of decreasing dimension
   starting with the highest-dimensional one.  All these routines have a
   last optional argument, which specifies a bound on the absolute value of
   integer coefficients in linear transformations.  For some massive integrals,
   it may be useful to set this to 2 instead of the default value of 1.

*)

Print["Barnes Routines, v 1.0.0 of November 1, 2007"];

(* Need to repair FindInstance in version 6.0; otherwise it will effectively
   hang when multiple points are requested *)
If [$VersionNumber >= 6,
   SetSystemOptions["ReduceOptions"->{"RandomInstances"->True}]];

BeginPackage["BarnesRoutines`"];

DoAllBarnes::usage = \
"DoAllBarnes[intlist_List,progressive_:False,bound_:1] repeatedly applies \
the Barnes lemmas to a list of integrals.  It first applies the second \
Barnes lemma, then the first Barnes lemma.  If the second argument is True, \
the lemmas are progressively applied to integrals of decreasing dimension \
starting with the highest-dimensional ones.  The last argument is passed to \
DoBarnes1 and DoBarnes2.";

DoBarnes1::usage = \
"DoBarnes1[int_MBint,bound_:1] applies the first Barnes lemma to the given \
MB integral.  It first looks for a linear transformation of the integration \
variables which will make such an application possible. The optional argument \
specifies a bound on the integer coefficients in such a transformation.";

DoBarnes2::usage = \
"DoBarnes2[int_MBint,bound_:1] applies the second Barnes lemma to the given \
MB integral.  It first looks for a linear transformation of the integration \
variables which will make such an application possible. The optional argument \
specifies a bound on the integer coefficients in such a transformation.";

ExpandInt::usage = \
"ExpandInt[int_MBint] expands the integrand, and then replaces the integral \
of a sum by a list of MB integrals.";

SetVerbosity::usage = \
"SetVerbosity[n_Integer] sets the reporting level for Barnes routines. \
Levels run from 0 to 3, and routines report more information at higher levels.";

(* Barnes1; Barnes2; conflict with MB's... *)

Begin["`Private`"];

MBint = MB`MBint;
MBmerge = MB`MBmerge;
MBcontinue = MB`MBcontinue;

verbosity = 0;
SetVerbosity[v_Integer] := (verbosity = v);

ShiftContours[integral_MBint,targetContour_List] :=
   (* Returns a list of integrals whose sum equals the original integral,
      where the original integral has its contours shifted according to
      the arguments.  The first integral is the shifted one (whose contours
      are then a linear approximation to an optimal one.)  The additional 
      integrals are residue terms in one of the variables.  *)
  Module[{integrand=integral[[1]],sourceContour=integral[[2,2]],vars,
          source,target,sign,perturb,perturbedInt},
    vars = #[[1]]& /@ sourceContour;
    If [(#[[1]]& /@ targetContour) =!= vars,
       Print["Variable mismatch! ",vars," vs ",(#[[1]]& /@ targetContour)];
       Return[{integral}]];
    added={};
    currentContour = sourceContour;
    Do[(* Do we cross poles going from the source to the destination contour? *)
       source = sourceContour[[j,2]];  target = targetContour[[j,2]];
       poles = 
          FindPoles[integrand,vars[[j]],Min[source,target],
                    Max[source,target],currentContour];
       sign = If [source < target,-1,1];
       If [verbosity >= 3,
          WriteString[$Output,"var: ",vars[[j]],"; poles: ",poles]];
       addedInt = sign (Plus @@ ((blip; FResidue[integrand,#])& /@ poles));
       (* It can happen that individual poles are singular though
          the expression as a whole is finite on the contour base;
          if that happens, perturb the contour slightly, and check that
          no new poles are crossed *)
       If [BadQ[addedInt /. currentContour],
           If [BadQ[MLimit[addedInt,currentContour]],
              Print["Warning: added integral singular on contour!"];
              Return[{integral}]];
           (* spurious singularity: perturb *)
           perturb = Array[sourceContour[[#,1]] -> 
                                sourceContour[[#,2]]+1/Prime[100+#]&,
                              Length[sourceContour]];
           If [Length[perturbedInt=ShiftContours[integral,perturb]] > 1,
               Print["Poles found in perturbation!"];
               Return[{integral}]];
           Print["Perturbing and re-shifting"];
           Return[ShiftContours[perturbedInt[[1]],targetContour]];
           ];
       If [verbosity >= 3,Print[""]];
       currentContour = Join[targetContour[[Range[1,j]]],
                           sourceContour[[Range[j+1,Length[sourceContour]]]]];
       If [addedInt == 0,Continue[]];
       added = Join[added,{MBint[addedInt,{{ep->0},
                     Join[targetContour[[Range[1,j-1]]],
                        sourceContour[[Range[j+1,Length[sourceContour]]]]]}]}];
       ,{j,1,Length[vars]}];
    Join[{MBint[integrand,{{ep->0},targetContour}]},added]
    ];

Perturb[integral_MBint] := 
  Module[{perturb,contour=integral[[2,2]]},
     perturb = Array[contour[[#,1]] -> contour[[#,2]]+1/Prime[100+#]&,
                     Length[contour]];
     If [Length[CleanInt[perturbedInt=ShiftContours[integral,perturb]]] > 1,
        Print["Poles found in perturbation!"];
        Return[integral]];
     perturbedInt[[1]]];

BadQ[v_] := MemberQ[{ComplexInfinity,Indeterminate,Infinity,-Infinity},v];

NaivelyBadIntQ[int_MBint] := BadQ[int[[1]] /. int[[2,2]]];
BadIntQ[int_MBint] := BadQ[MLimit[int[[1]],int[[2,2]]]];

(* Replace variable in contour *)
ReplaceContourVar[contour_,var_,newvar_,
                  rep_ (* Expression in terms of old vars *)] :=
   Module[{pos=Flatten[Position[contour[[All,1]],var]][[1]]},
      Join[contour[[Range[pos-1]]],{newvar -> (rep /. contour)},
           contour[[Range[pos+1,Length[contour]]]]]];
            
(* Replace "var" by the expression "rep" (which should be var+x), that
   is "var" is the original variable *)
ReplaceVar2[integral_MBint,var_ -> rep_] :=
  Module[{newvar,newrep},
    newrep = newvar /. Flatten[Solve[var == (rep /. var->newvar),newvar]];
    MBint[integral[[1]] /. var -> rep,
      {integral[[2,1]],
       ReplaceContourVar[integral[[2,2]],var,newvar,newrep]}] /.
      newvar -> var];

ReplaceAllVars[integral_MBint,reps_List] :=
  Module[{result=integral},
    Scan[(result = ReplaceVar2[result,#])&,reps];  result];

ExpandInt[int_MBint] := 
  MBint[#,Sequence @@ Rest[int]]& /@ (List @@ Expand[int[[1]]]) /; 
     (* 10/2/06 Need to expand things that start out as sums too! 
        int[[1]] =!= Expand[int[[1]]] && *) Head[Expand[int[[1]]]] === Plus;
ExpandInt[int_MBint] := int;

ExpandArgs[expr_] := expr /. a_^n_ :> a^Expand[n] /. {
  Gamma[a_] :> Gamma[Expand[a]],
  PolyGamma[n_, a_] :> PolyGamma[n, Expand[a]]};

MBDimension[int_MBint] := Length[int[[2,2]]];
MellinVars[int_MBint] := First /@ int[[2,2]];

SummarizeDimensions[ints_List] :=
  Module[{dims=(MBDimension /@ ints)},
    ToString[Count[dims,#]]<>" x "<>ToString[#]& /@
               Reverse[Join[{0},Range[Max @@ dims]]]
    ];

(* Simplified version of DoAllBarnesIntermixedWithMerging, M.C. *)

DoAllBarnes[intlist_List,progressive_:False,bound_:1] :=
  (* "progressive" means apply only to integrals above a limit.  The limit
     starts out as the maximal dimension, and is lowered by one on each
     pass *)
  Module[{result,previous={},limit=1,others},
    If [verbosity >= 1, WriteString[$Output,"[",Length[intlist],"]"]];
    If [verbosity >= 2, Print[""]];
    result = Flatten[ExpandInt /@ intlist];
    If [progressive, limit = Max @@ (MBDimension /@ result)];
    (* Note that SortForMerging does not produce a canonical order... *)
    While[Union[result] =!= Union[previous],
       previous = result;
       Print[SummarizeDimensions[result]];
       (* Barnes2 *)
       If [verbosity >= 2, Print["Barnes 2"]];
       others = Select[result,MBDimension[#] < limit&];
       result = Select[result,MBDimension[#] >= limit&];
       result = Flatten[DoBarnes2 /@ result];
       result = Flatten[ExpandInt /@ result];
       result = Join[result,others];
       result = MBmerge[result];
       If [verbosity >= 1, WriteString[$Output,"[",Length[result],"]"]];
       result = Flatten[ExpandInt /@ result];
       (* Barnes1 *)
       If [verbosity >= 2, Print["Barnes 1"]];
       others = Select[result,MBDimension[#] < limit&];
       result = Select[result,MBDimension[#] >= limit&];
       result = Flatten[DoBarnes1 /@ result];
       result = Flatten[ExpandInt /@ result];
       result = Join[result,others];
       result = MBmerge[result];
       If [verbosity >= 1, WriteString[$Output,"[",Length[result],"]"]];
       result = Flatten[ExpandInt /@ result];
       If [Head[interfile] === String,
         (* interfile should have been defined by our caller... *)
         xresult = result;
         Save[interfile<>".m",limit,xresult];  Clear[xresult]];
       If [limit > 1, limit -= 1]];
    result];

(* Find a linear transformation of the variables that allows 
   the first Barnes lemma to be applied; the integrand must be a
   pure product.  It returns a list {var,replacements} specifying
   the variable to which the 1st Barnes lemma should be applied after
   performing the replacements using ReplaceVar2 (or an empty list if
   no solution is found).  The list of replacements may be empty of course.
   10/3/06: Look for transformation-free solution (i.e. just plain variable)
            preferentially.
 *)
FindBarnesable[int_MBint,bound_:1] :=
  Module[{integrand=int[[1]],vars=#[[1]]& /@ int[[2,2]],
          numer,denom,transform,rep,args,constraint,avars,j,dvars,
          zrep,c,backup={},backupZrep = {},backupVar,allsols},
     (* Avoid errors that would otherwise arise... *)
     If [NumberQ[integrand],Return[{}]];
     If [integrand =!= Expand[integrand] || Head[integrand] =!= Times,
        Return[{}]];
     (* The code below won't really work in one-dimensional cases, so
        ignore this *)
     If [Length[vars]<=1,Return[{}]];
     numer = Numerator[integrand];
     denom = Denominator[integrand];
     (* 9/29/06 Be careful to treat T^(z1-z2) as a single exponent,
                even though one will appear in the numerator, the 
                other in the denominator; we can't use Variables to
                extract this properly *)
     powers = 1;
     If [Head[numer]===Times,
        powers = Select[numer,Head[#]===Power && 
              (MemberQ[{Symbol,Integer,Real},Head[First[#]]] || 
               MatchQ[First[#],_Symbol[_]])&]];
     If [Head[denom]===Times,
        powers /= Select[denom,Head[#]===Power && Head[#[[1]]]===Symbol&]];
     (* The basic approach is to require the variable (after limited
        linear changes) to be present in only four gamma functions,
        only in the numerator, with (++--) coefficients. *)
     (* Need to keep track of multiplicity in numerator *)
     numer = List @@ numer;
     (* Lone powers of PolyGamma[0,x] can be treated if x matches one of
        the Gamma functions; the matching to the Gammas is enforced below  *)
     allowedNumer = Select[numer,MatchQ[#,PolyGamma[0,_]]&];
     numer = numer /. 
             f_[x__]^p_ :> Array[f[x]&,p] /; MemberQ[{Gamma,PolyGamma},f];
     numer = Flatten[numer];
     otherNumer = Select[numer,MemberQ[{PolyGamma},Head[#]]&];
     (* Let var appear in PolyGamma[0,_], so long as the PolyGamma
        appears only in the first power, and in the same way as in one
        of the Gammas; otherwise, it's a forbidden form, which we tack
        on to the "denom" set. *)
     allowedNumer = Select[allowedNumer,MemberQ[numer,Gamma[Last[#]]]&];
     otherNumer = Complement[otherNumer,allowedNumer];
     numer = Select[numer,MemberQ[{Gamma},Head[#]]&];
     denom = Select[Variables[denom],MemberQ[{Gamma,PolyGamma},Head[#]]&];
     (* Extract arguments of these special functions *)
     numer = Last /@ numer;
     denom = Last /@ denom;
     If [Head[powers] === Times,
        powers = Last /@ (List @@ powers)];
     If [Head[powers] === Power,
        powers = {Last[powers]}];
     If [Head[powers] === Integer,
        powers = {}];
     allowedNumer = Last /@ allowedNumer;
     (* If something is both in "allowed" and "other" (e.g. PolyGamma[0,x]
        & PolyGamma[1,x] appear), it's not allowed... *)
     otherNumer = Last /@ otherNumer;
     allowedNumer = Complement[allowedNumer,otherNumer];
     denom = Join[denom,powers];
     denom = Join[denom,otherNumer];
     transform = Array[c,{Length[vars],Length[vars]}];
     (* Diagonal elements 1 *)
     transform = transform /. c[i_,i_]->1;
     rep = Thread[vars->transform.vars];
     args = D[numer /. rep,#]& /@ vars;
     dargs = D[denom /. rep,#]& /@ vars;
     dvars = Variables /@ dargs;
     If [verbosity >= 2, WriteString[$Output,"Looking at "]];
     Do[
       If [verbosity >= 2, WriteString[$Output,vars[[j]]]];
       (* Look for simplest changes of variables *)
       avars = Variables[args[[j]]];
       constraint = And @@ (-bound <= # <= bound & /@ avars);
       constraint = constraint && (And @@ (# == 0& /@ dargs[[j]]));
       eqns = args[[j]].args[[j]] == 4 \
                   && (Plus @@ args[[j]]) == 0&& constraint;
       sol=Flatten[FindInstance[eqns,Join[avars,dvars[[j]]],Integers]];
       (* Prune down to either modified variables, or target variable *)
       zrep = Select[rep /. sol /. c[__]->0,#[[1]]=!=#[[2]]&];
       If [Length[sol] > 0,
           If [verbosity >= 2, WriteString[$Output,"."];
               Print[" ",If [Length[zrep] > 0,InputForm[zrep],
                         "[No change of variable needed]"]]];
           Return[{vars[[j]],zrep},Module]];
       If [Length[sol] > 0 && Length[zrep] == 0,
           If [verbosity >= 2, Print[". [No change of variable needed]"]];
           Return[{vars[[j]],zrep},Module]];
       (* Otherwise file it away *)
       If [Length[sol] > 0 && Length[backup] == 0, 
           If [verbosity >= 2, WriteString[$Output,"!"]];
           backup = sol;  backupZrep = zrep;  backupVar = vars[[j]],
           If [verbosity >= 2, WriteString[$Output,";"]]];
       ,{j,1,Length[vars]}];
     (* No simple solution? Return back-up solution if any *)
     If [Length[backup] > 0,
         If [verbosity >= 2, Print[". ",InputForm[backupZrep]]];
         Return[{backupVar,backupZrep},Module]];
     If [verbosity >= 2, Print[" solution not found"]];
     {}
     ];

(* We use weak (>=, <=) inequalities & not just strong (>,<) ones
   in FindBarnesable & FindBarnesable2 (as well as elsewhere); this
   should make FindInstance run much faster for them (according to 
     forums.wolfram.com/mathgroup/archive/2006/Jan/msg00387.html)
*)

Developer`SetSystemOptions["InequalitySolvingOptions"->{"BrownProjection"->False}];

FBtimeLimit = 60;
(* 9/25/06: Same thing for second Barnes lemma *)
(* 9/29/06: this routine sometimes seems to hang or at least be very very
            slow in Mathematica 5.0;
            seems to work fine in Mathematica 5.2 *)
FindBarnesable2[int_MBint,bound_:1] :=
  Module[{integrand=int[[1]],vars=#[[1]]& /@ int[[2,2]],
          numer,denom,transform,rep,args,constraint,avars,j,dvars,
          zrep,p,c,backup={},backupZrep = {},backupVar},
     (* Avoid errors that would otherwise arise... *)
     If [NumberQ[integrand],Return[{}]];
     (* The code below won't really work in one-dimensional cases, so
        ignore this *)
     If [Length[vars]<=1,Return[{}]];
     If [integrand =!= Expand[integrand] || Head[integrand] =!= Times,
        Return[{}]];
     numer = Numerator[integrand];
     denom = Denominator[integrand];
     powers = 1;
     If [Head[numer]===Times,
        powers = Select[numer,Head[#]===Power && 
              (MemberQ[{Symbol,Integer,Real},Head[First[#]]] || 
               MatchQ[First[#],_Symbol[_]])&]];
     If [Head[denom]===Times,
        powers /= Select[denom,Head[#]===Power && Head[#[[1]]]===Symbol&]];
     (* The basic approach is to require the variable (after limited
        linear changes) to be present in exactly five numerator gamma 
        functions and one denominator function, with either (+++--;+)
        or (---++;-) coefficients. *)
     (* Need to keep track of multiplicity in numerator *)
     numer = List @@ numer;
     numer = numer /. 
             f_[x__]^p_ :> Array[f[x]&,p] /; MemberQ[{Gamma,PolyGamma},f];
     numer = Flatten[numer];
     otherNumer = Select[numer,MemberQ[{PolyGamma},Head[#]]&];
     numer = Select[numer,MemberQ[{Gamma},Head[#]]&];
     denom = Select[Variables[denom],MemberQ[{Gamma,PolyGamma},Head[#]]&];
     (* Don't allow PolyGammas for the moment *)
     disallow = otherNumer;
     (* Extract arguments of these special functions *)
     numer = Last /@ numer;
     denom = Last /@ denom;
     If [Head[powers] === Times,
        powers = Last /@ (List @@ powers)];
     If [Head[powers] === Power,
        powers = {Last[powers]}];
     If [Head[powers] === Integer,
        powers = {}];
     disallow = Last /@ disallow;
     denom = Join[denom];
     disallow = Join[disallow,powers];
     transform = Array[c,{Length[vars],Length[vars]}];
     (* Diagonal elements 1 *)
     transform = transform /. c[i_,i_]->1;
     rep = Thread[vars->transform.vars];
     args = D[numer /. rep,#]& /@ vars;
     dargs = D[denom /. rep,#]& /@ vars;
     disargs = D[disallow /. rep,#]& /@ vars;
     dvars = Variables /@ dargs;
     If [verbosity >= 2, WriteString[$Output,"Looking at "]];
     Do[
       If [verbosity >= 2, WriteString[$Output,vars[[j]]]];
       (* Look for simplest changes of variables *)
       avars = Variables[args[[j]]];
       constraint = And @@ (-bound <= # <= bound & /@ avars);
       constraint = constraint && (And @@ (# == 0& /@ disargs[[j]]));
       (* There's an additional constraint for residual == -1:
            one numerator argument has to be one less than the denominator 
            argument (that is we must have 
                  G[a+z] G[b+z] G[c-z] G[2-a-b-c-z] G[e-z]/G[1+e-z])
          to do it in one go we introduce a selector variable p,
                  sum p_i = 1, and 0 <= p_i <= 1,
          and require that (p*coeffs).numer - denom = -1,
          along with the constraint that p.coeffs = 1 or -1 (according
          to the sum of numerator coeffs),
          so that p_i corresponds to a "live" numerator *)
       pvars = Array[p,Length[args[[j]]]];
       pConstraint = And @@ (0 <= # <= 1 & /@ pvars);
       pConstraint = (Plus @@ pvars == 1) && pConstraint;
       (* Residuals should be summed over all terms where the var
          is present after the transformation, with sign independent
          of the sign of the var inside the term -- square the coefficient
          (which ultimately will be +/- 1) to get that *)
       residual = denom.(dargs[[j]]^2) - numer.(args[[j]]^2);
       (* Fix other variables to their contour values to "distinct" values
          to avoid problems with non-constant expressions in FindInstance;
          they must of course cancel if we're to find a solution *)
       residual = residual /. vars[[j]]->0;
       residual = residual /. Array[(vars[[#]]->1/Prime[100+#])&,Length[vars]];       xresidual = Expand[residual];
       (* This picks out only terms which actually have the variable --
          but multiplied by the sign in the denominator = overall sign
          in numerator; take that into account below... *)
       difference = (pvars args[[j]]).numer-denom.dargs[[j]];
       difference = difference /. vars[[j]]->0;
       difference = difference /. Array[(vars[[#]]->1/Prime[100+#])&,Length[vars]];       xdifference = Expand[difference];
       eqns = args[[j]].args[[j]] == 5 \
                   && dargs[[j]].dargs[[j]] == 1 \
                   && ((Plus @@ args[[j]] == 1 && Plus @@ dargs[[j]] == 1) ||
                       (Plus @@ args[[j]] == -1 && Plus @@ dargs[[j]] == -1)) \
                   && (residual == 0 || residual == 1 || residual == 2 
                      || (residual == -1 && difference == -(Plus @@ dargs[[j]])
                           && pvars.args[[j]] == Plus @@ args[[j]]
                           && pConstraint)) \
                   && constraint;
       (* To speed things up, first look for a solution without
          changes of variable; we want such solutions preferentially
          anyway *)
       If [pass == 1,eqns = eqns /. {c[__]->0}];
       (* In higher dimensions, it appears to hang; limit the amount of
          time to stop that... *)
       sol = TimeConstrained[Flatten[FindInstance[eqns,Join[avars,dvars[[j]],
                                             pvars],Integers]],FBtimeLimit,{}];
       If [!FreeQ[sol,FindInstance],Throw["oops!"]];
       (* Prune down either to modified variables, or target variable *)
       zrep = Select[rep /. sol /. c[__]->0,#[[1]]=!=#[[2]]&];
       If [Length[sol] > 0,
          If [verbosity >= 2, WriteString[$Output,"."];
              Print[" ",If [Length[zrep] > 0,InputForm[zrep],
                         "[No change of variable needed]"]]];
           Return[{vars[[j]],zrep},Module]];
       If [verbosity >= 2, WriteString[$Output,";"]];
       If [Length[sol] > 0 && Length[zrep] == 0,
           If [verbosity >= 2, Print[". [No change of variable needed]"]];
           Return[{vars[[j]],zrep},Module]];
       (* With multiple passes, no need to file it away -- just return it *)
       If [Length[sol] > 0 && pass > 1, 
           If [verbosity >= 2, Print[". ",InputForm[zrep]]];
           Return[{vars[[j]],zrep},Module]];
       If [verbosity >= 2, WriteString[$Output,";"]];
       ,{pass,1,2},{j,1,Length[vars]}];
     (* No simple solution? Return back-up solution if any
        Not needed with multiple passes as above *)
     If [verbosity >= 2, Print[" solution not found"]];
     {}
     ];

(* Replacements which can increase the chances of applying the Barnes
   lemmas 

   10/14/06: Tightened conditions on powers
*)
simplifyForBarnes = {
  Gamma[mz_]^p1_. Gamma[1+z_]^p2_./Gamma[1+mz_] :> 
        -Gamma[-z]^(p1-1) Gamma[1+z]^(p2-1) Gamma[z] /; Expand[z+mz]==0,
  Gamma[mz_]^p_. Gamma[1+z_]/Gamma[1+mz_]^2 :>
        -Gamma[-z]^(p-1) Gamma[z]/Gamma[1-z] /; Expand[z+mz]==0,
  Gamma[mz_]^p_. Gamma[1+z_]/Gamma[-1+mz_] :> 
        -Gamma[-z]^(p-1) Gamma[2+z] /; Expand[z+mz]==0,
  Gamma[-1+mz_]^p1_. Gamma[2+z_]^p2_./Gamma[mz_] :> 
        -Gamma[-1-z]^(p1-1) Gamma[2+z]^(p2-1) Gamma[1+z] /; 
             Expand[z+mz]==0 && p1>0 && p2>0,
  Gamma[-1+mz_]^p_. Gamma[2+z_]/Gamma[mz_]^2 :> 
        -Gamma[-1-z]^(p-1) Gamma[1+z]/Gamma[-z] /; Expand[z+mz]==0 && p>0,
  Gamma[-1+mz_]^p_. Gamma[2+z_]/Gamma[1+z_] :> 
        -Gamma[mz] Gamma[-1+mz]^(p-1) /; Expand[z+mz]==0 && p>0,
  (* Added 10/14/06 *)
  (Gamma[-1+mz_] Gamma[mz_] Gamma[2+z_])/Gamma[1+mz_] :> 
        Gamma[-z]*Gamma[z] /; Expand[z+mz]==0,
  (* Be careful if we have both PolyGamma[0,1-z] & PolyGamma[0,1+z];
     that would cause simplifyForBarnes to oscillate back & forth between
     the two equivalent representations *)
  Gamma[mz_] Gamma[1+z_] PolyGamma[0,1+mz_] :> 
        -Gamma[1-z] Gamma[z] PolyGamma[0,1-z] /; 
               Expand[z+mz]==0 && Head[mz] === Times,
  Gamma[mz_] Gamma[1+z_]^p_./Gamma[2+z_] :> 
        -Gamma[-1-z] Gamma[1+z]^(p-1) /; Expand[z+mz]==0,
  Gamma[mz_] Gamma[1+z_]^p_./Gamma[2+z_]^2 :> 
        -Gamma[-1-z] Gamma[1+z]^(p-1)/Gamma[2+z]  /; Expand[z+mz]==0,
  Gamma[-1+mz_] Gamma[1+z_] Gamma[2+z_]/Gamma[mz_] :> 
        -Gamma[1+z]^2 /; Expand[z+mz]==0,
  Gamma[1+mz_] Gamma[mz_] Gamma[z_]/Gamma[1+z_] :> 
        -Gamma[-z]^2 /; Expand[z+mz]==0,
  Gamma[-2+mz_]^p1_. Gamma[3+z_]^p2_./Gamma[-1+mz_] :>
        -Gamma[2+z] Gamma[-2-z]^(p1-1) Gamma[3+z]^(p2-1) /; Expand[z+mz]==0,
  Gamma[1+za_] Gamma[mza_]^p_./Gamma[1+mza_] :> 
     -Gamma[za] Gamma[mza]^(p-1) /; Expand[za+mza]==0,
  Gamma[-1-za_-zb_] Gamma[2+za_+zb_]/Gamma[-za_-zb_] -> -Gamma[1+za+zb],
  Gamma[-za_-zb_] Gamma[1+za_+zb_]^p_./Gamma[2+za_+zb_] ->
               -Gamma[-1-za-zb] Gamma[1+za+zb],
  Gamma[za_-zb_]^2 Gamma[1-za_+zb_]/Gamma[1+za_-zb_] -> 
       -Gamma[za-zb] Gamma[zb-za],
  Gamma[-1-za_-zb_] Gamma[2+za_+zb_] PolyGamma[0,1+za_+zb_] ->
     -Gamma[-za-zb] Gamma[1+za+zb] PolyGamma[0,1+za+zb],
  Gamma[-1-za_-zb_] Gamma[2+za_+zb_] PolyGamma[0,-za_-zb_] ->
     -Gamma[-za-zb] Gamma[1+za+zb] PolyGamma[0,-za-zb]};

  (* The following sometimes blocks Barnes, if a PolyGamma is present...
     and the FreeQ's don't seem to accomplish what they're supposed to,
     so drop it for the moment 
     Should it in any case be done only if -1-z1-z2 < 0? *)
  Gamma[-1-z1_-z2_] Gamma[2+z1_+z2_] rest_?(FreeQ[#,PolyGamma[0,2+z1_+z2_]]&) :> 
     -Gamma[-z1-z2] Gamma[1+z1+z2] rest /; 
       FreeQ[rest,PolyGamma[0,-1-z1-z2]] && FreeQ[rest,PolyGamma[0,2+z1+z2]];

(* Using the Barnes lemma may result in integrals which are naively
   divergent -- separate terms are independently divergent on the 
   contour, but the poles actually cancel between terms.  We assume
   they're not genuinely divergent -- that will anyhow eventually be
   picked up by perturbing code.  To repair this we perturb the 
   contours *)
RepairNaivelyBad[int_MBint] :=
  If [NaivelyBadIntQ[int],
(*     Print["Warning: generated naively divergent integral!"];*)
     xtarget = int;
     file = ToString[Unique[xtag]];
     WriteString[$Output," [",file,"]"];
     Save[file<>".m",xtarget];
     If [BadIntQ[int],
         Print["Oops! Generated genuinely divergent integral"];
         Throw[DivergentIntegral]];
     Perturb[int],int];

DoBarnes1[int_MBint,bound_:1] :=
  Module[{sol,result,anySol = False, report = True},
   result = RepairNaivelyBad /@ (
   (sol = FindBarnesable[# //. simplifyForBarnes,bound];
   If [Length[sol]>0, anySol = True; 
       If [verbosity >= 1, WriteString[$Output,"Doing ",sol[[1]]]];
           currentTarget = #;
          Barnes1[ExpandArgs[ReplaceAllVars[# //. simplifyForBarnes,sol[[2]]]],
                            sol[[1]]],
       If [MBDimension[#]==1,
           If [verbosity >= 1 && report, report = False;
               WriteString[$Output,"Trying ",First[MellinVars[#]]]];
           Barnes1[#,First[MellinVars[#]]],#]]
   )& /@ Flatten[{ExpandInt[int]}]);
   result = ExpandArgs[result];
   If [verbosity >= 1 && MBDimension[int] == 1,
      If [Union[MBDimension /@ result] =!= {1},
          Print["; ",{MBDimension[int]} -> (MBDimension /@ result)];,
          Print["; failed"]]];
   If [anySol,
        If [verbosity >= 1, 
            Print["; ",{MBDimension[int]} -> (MBDimension /@ result)]];
      (*If [{MBDimension[int]} == (MBDimension /@ result),
           xsource = int; xres = result; Throw["missed?"]]*)];
   CheckRes /@ result;
   result];

DoBarnes2[int_MBint,bound_:1] :=
  Module[{sol,result,anySol = False, report = True},
   result = RepairNaivelyBad /@ (
   (sol = FindBarnesable2[# //. simplifyForBarnes,bound];
   If [Length[sol]>0, anySol = True;
       If [verbosity >= 1, WriteString[$Output,"Doing ",sol[[1]]]];
           currentTarget = #;
          Barnes2[ExpandArgs[ReplaceAllVars[# //. simplifyForBarnes,sol[[2]]]],
                  sol[[1]]],
       If [MBDimension[#]==1,
           If [verbosity >= 1 && report, report = False;
               WriteString[$Output,"Trying ",First[MellinVars[#]]]];
           Barnes2[#,First[MellinVars[#]]],#]]
   )& /@ Flatten[{ExpandInt[int]}]);
   result = ExpandArgs[result];
    If [True,
      bad = Select[result,NaivelyBadIntQ];
      If [Length[bad] > 0,
         If [verbosity >= 1,
            Print["B2: ",Length[bad]," naively divergent integrals"]];
         bad = Select[bad,BadIntQ];
         Print["B2: ",Length[bad]," genuinely divergent integrals"]]];
   If [verbosity >= 1 && MBDimension[int] == 1,
      If [Union[MBDimension /@ result] =!= {1},
          Print["; ",{MBDimension[int]} -> (MBDimension /@ result)];,
          Print["; failed"]]];   
   If [anySol,
       If [verbosity >= 1,
          Print["; ",{MBDimension[int]} -> (MBDimension /@ result)]];
      (*If [{MBDimension[int]} == (MBDimension /@ result),
           xsource = int; xres = result; Throw["missed?"]]*)];
   CheckRes /@ result;
   result];

debugBarnes = False;
verboseBarnes = False;
searchTimeLimit = 60;
recordFailedBarnes = False;
failedBarnes1 = {};
failedBarnes2 = {};

CanonicalGamma[expr_,vars_List] :=
   expr //. {Gamma[x_] :> Gamma[x+1]/x /; (x /. (#->0& /@ vars)) < 1,
             Gamma[x_] :> Gamma[x-1] (x-1) /; (x /. (#->0& /@ vars)) > 1,
             PolyGamma[j_,x_] :> PolyGamma[j,1+x]-(-1)^j j!/x^(j+1) /; 
                                        (x /. (#->0& /@ vars)) < 1,
             PolyGamma[j_,x_] :> PolyGamma[j,x-1]+(-1)^j j!/(x-1)^(j+1) /; 
                                        (x /. (#->0& /@ vars)) > 1
              };

Barnes1[MBint[integrand_, {fixedVarRules_List, intVarRules_List}], z_Symbol] := 
Block[{original = MBint[integrand, {fixedVarRules, intVarRules}],
  rules = Flatten[{fixedVarRules, intVarRules}],
  integ = integrand,
  x, pos, a, b, p, neg, c, d, n, xx, xx0, int, fixed, cont, ct, orig,
  patterns, pattern, continuation, continue, shift},

  (* Modified 9/24/06 (DAK) to match forms with PolyGamma[0,ai+z] in them *)
  (* "patterns" is a list of pattern->replacement entries; note that
     the values of "a", "b", "c", "d" are determined below.  *)
  patterns = {
     Gamma[a + z]*Gamma[b + z]*Gamma[c - z]*Gamma[d - z] ->
      Gamma[a + c]*Gamma[a + d]*Gamma[b + c]*Gamma[b + d]/Gamma[a + b + c + d],
     (* Single PolyGamma with matching argument *)
     Gamma[a + z]*Gamma[b + z]*Gamma[c - z]*Gamma[d - z] PolyGamma[0,a+z] ->
      Gamma[a + c]*Gamma[a + d]*Gamma[b + c]*Gamma[b + d]*
          (PolyGamma[0,a+c]+PolyGamma[0,a+d]-PolyGamma[0,a+b+c+d])/
              Gamma[a + b + c + d],
     Gamma[a + z]*Gamma[b + z]*Gamma[c - z]*Gamma[d - z] PolyGamma[0,b+z] ->
      Gamma[a + c]*Gamma[a + d]*Gamma[b + c]*Gamma[b + d]*
          (PolyGamma[0,b+c]+PolyGamma[0,b+d]-PolyGamma[0,a+b+c+d])/
              Gamma[a + b + c + d],
     Gamma[a + z]*Gamma[b + z]*Gamma[c - z]*Gamma[d - z] PolyGamma[0,c-z] ->
      Gamma[a + c]*Gamma[a + d]*Gamma[b + c]*Gamma[b + d]*
          (PolyGamma[0,a+c]+PolyGamma[0,b+c]-PolyGamma[0,a+b+c+d])/
              Gamma[a + b + c + d],
     Gamma[a + z]*Gamma[b + z]*Gamma[c - z]*Gamma[d - z] PolyGamma[0,d-z] ->
      Gamma[a + c]*Gamma[a + d]*Gamma[b + c]*Gamma[b + d]*
          (PolyGamma[0,a+d]+PolyGamma[0,b+d]-PolyGamma[0,a+b+c+d])/
              Gamma[a + b + c + d],
     (* Pairs of PolyGammas with matching argument *)
     Gamma[a + z]*Gamma[b + z]*Gamma[c - z]*Gamma[d - z] \
        PolyGamma[0,a+z] PolyGamma[0,b+z] ->
         Gamma[a + c]*Gamma[b + c]*Gamma[a + d]*Gamma[b + d]*
          (PolyGamma[0, a + c]*PolyGamma[0, b + c] + PolyGamma[0, b + c]*
           PolyGamma[0, a + d] + PolyGamma[0, a + c]*PolyGamma[0, b + d] +
           PolyGamma[0, a + d]*PolyGamma[0, b + d] - PolyGamma[0, a + c]*
           PolyGamma[0, a + b + c + d] - PolyGamma[0, b + c]*
           PolyGamma[0, a + b + c + d] - PolyGamma[0, a + d]*
           PolyGamma[0, a + b + c + d] - PolyGamma[0, b + d]*
           PolyGamma[0, a + b + c + d] + PolyGamma[0, a + b + c + d]^2 -
           PolyGamma[1, a + b + c + d])/Gamma[a + b + c + d],
     Gamma[a + z]*Gamma[b + z]*Gamma[c - z]*Gamma[d - z] \
        PolyGamma[0,a+z] PolyGamma[0,c-z] ->
         Gamma[a + c]*Gamma[b + c]*Gamma[a + d]*Gamma[b + d]*
          (PolyGamma[0, a + c]^2 + PolyGamma[0, a + c]*PolyGamma[0, b + c] +
           PolyGamma[0, a + c]*PolyGamma[0, a + d] + PolyGamma[0, b + c]*
           PolyGamma[0, a + d] - 
           2*PolyGamma[0, a + c]*PolyGamma[0, a + b + c + d] -
           PolyGamma[0, b + c]*PolyGamma[0, a + b + c + d] -
           PolyGamma[0, a + d]*PolyGamma[0, a + b + c + d] +
           PolyGamma[0, a + b + c + d]^2 + PolyGamma[1, a + c] -
           PolyGamma[1, a + b + c + d])/Gamma[a + b + c + d],
     Gamma[a + z]*Gamma[b + z]*Gamma[c - z]*Gamma[d - z] \
        PolyGamma[0,a+z] PolyGamma[0,d-z] ->
         Gamma[a + c]*Gamma[b + c]*Gamma[a + d]*Gamma[b + d]*
          (PolyGamma[0, a + c]*PolyGamma[0, a + d] + PolyGamma[0, a + d]^2 +
           PolyGamma[0, a + c]*PolyGamma[0, b + d] + PolyGamma[0, a + d]*
           PolyGamma[0, b + d] - 
           PolyGamma[0, a + c]*PolyGamma[0, a + b + c + d] -
           2*PolyGamma[0, a + d]*PolyGamma[0, a + b + c + d] -
           PolyGamma[0, b + d]*PolyGamma[0, a + b + c + d] +
           PolyGamma[0, a + b + c + d]^2 + PolyGamma[1, a + d] -
           PolyGamma[1, a + b + c + d])/Gamma[a + b + c + d],
     Gamma[a + z]*Gamma[b + z]*Gamma[c - z]*Gamma[d - z] \
        PolyGamma[0,b+z] PolyGamma[0,c-z] ->
         Gamma[a + c]*Gamma[b + c]*Gamma[a + d]*Gamma[b + d]*
          (PolyGamma[0, a + c]*PolyGamma[0, b + c] + PolyGamma[0, b + c]^2 +
           PolyGamma[0, a + c]*PolyGamma[0, b + d] + PolyGamma[0, b + c]*
           PolyGamma[0, b + d] - 
           PolyGamma[0, a + c]*PolyGamma[0, a + b + c + d] -
           2*PolyGamma[0, b + c]*PolyGamma[0, a + b + c + d] -
           PolyGamma[0, b + d]*PolyGamma[0, a + b + c + d] +
           PolyGamma[0, a + b + c + d]^2 + PolyGamma[1, b + c] -
           PolyGamma[1, a + b + c + d])/Gamma[a + b + c + d],
     Gamma[a + z]*Gamma[b + z]*Gamma[c - z]*Gamma[d - z] \
        PolyGamma[0,b+z] PolyGamma[0,d-z] ->
         Gamma[a + c]*Gamma[b + c]*Gamma[a + d]*Gamma[b + d]*
          (PolyGamma[0, b + c]*PolyGamma[0, a + d] + PolyGamma[0, b + c]*
           PolyGamma[0, b + d] + PolyGamma[0, a + d]*PolyGamma[0, b + d] +
           PolyGamma[0, b + d]^2 - 
           PolyGamma[0, b + c]*PolyGamma[0, a + b + c + d] -
           PolyGamma[0, a + d]*PolyGamma[0, a + b + c + d] -
           2*PolyGamma[0, b + d]*PolyGamma[0, a + b + c + d] +
           PolyGamma[0, a + b + c + d]^2 + PolyGamma[1, b + d] -
           PolyGamma[1, a + b + c + d])/Gamma[a + b + c + d],
     Gamma[a + z]*Gamma[b + z]*Gamma[c - z]*Gamma[d - z] \
        PolyGamma[0,c-z] PolyGamma[0,d-z] ->
         Gamma[a + c]*Gamma[b + c]*Gamma[a + d]*Gamma[b + d]*
          (PolyGamma[0, a + c]*PolyGamma[0, a + d] + PolyGamma[0, b + c]*
           PolyGamma[0, a + d] + PolyGamma[0, a + c]*PolyGamma[0, b + d] +
           PolyGamma[0, b + c]*PolyGamma[0, b + d] - PolyGamma[0, a + c]*
           PolyGamma[0, a + b + c + d] - PolyGamma[0, b + c]*
           PolyGamma[0, a + b + c + d] - PolyGamma[0, a + d]*
           PolyGamma[0, a + b + c + d] - PolyGamma[0, b + d]*
           PolyGamma[0, a + b + c + d] + PolyGamma[0, a + b + c + d]^2 -
           PolyGamma[1, a + b + c + d])/Gamma[a + b + c + d],
      (* 9/30/06 Triplets of PolyGammas with different arguments *)
     Gamma[a + z]*Gamma[b + z]*Gamma[c - z]*Gamma[d - z] \
        PolyGamma[0,a+z] PolyGamma[0,b+z] PolyGamma[0,c-z] ->
         (Gamma[a + c]*Gamma[b + c]*Gamma[a + d]*Gamma[b + d]*
           (PolyGamma[0, a + c]^2*PolyGamma[0, b + c] + PolyGamma[0, a + c]*
             PolyGamma[0, b+c]^2+PolyGamma[0, a+c]*PolyGamma[0, b+c]*
             PolyGamma[0, a+d]+PolyGamma[0, b+c]^2*PolyGamma[0, a+d] +
            PolyGamma[0, a+c]^2*PolyGamma[0, b+d]+PolyGamma[0, a+c]*
             PolyGamma[0, b+c]*PolyGamma[0, b+d]+PolyGamma[0, a+c]*
             PolyGamma[0, a+d]*PolyGamma[0, b+d]+PolyGamma[0, b+c]*
             PolyGamma[0, a+d]*PolyGamma[0, b+d] - PolyGamma[0, a+c]^2*
             PolyGamma[0, a+b+c+d] - 3*PolyGamma[0, a+c]*PolyGamma[0, b+c]*
             PolyGamma[0, a+b+c+d] - PolyGamma[0, b+c]^2*
             PolyGamma[0, a+b+c+d] - PolyGamma[0, a+c]*PolyGamma[0, a+d]*
             PolyGamma[0, a+b+c+d] - 2*PolyGamma[0, b+c]*PolyGamma[0, a+d]*
             PolyGamma[0, a+b+c+d] - 2*PolyGamma[0, a+c]*PolyGamma[0, b+d]*
             PolyGamma[0, a+b+c+d] - PolyGamma[0, b+c]*PolyGamma[0, b+d]*
             PolyGamma[0, a+b+c+d] - PolyGamma[0, a+d]*PolyGamma[0, b+d]*
             PolyGamma[0, a+b+c+d]+2*PolyGamma[0, a+c]*
             PolyGamma[0, a+b+c+d]^2+2*PolyGamma[0, b+c]*
             PolyGamma[0, a+b+c+d]^2+PolyGamma[0, a+d]*
             PolyGamma[0, a+b+c+d]^2+PolyGamma[0, b+d]*
             PolyGamma[0, a+b+c+d]^2 - PolyGamma[0, a+b+c+d]^3 +
            PolyGamma[0, b+c]*PolyGamma[1, a+c]+PolyGamma[0, b+d]*
             PolyGamma[1, a+c] - PolyGamma[0, a+b+c+d]*PolyGamma[1, a+c] +
            PolyGamma[0, a+c]*PolyGamma[1, b+c]+PolyGamma[0, a+d]*
             PolyGamma[1, b+c] - PolyGamma[0, a+b+c+d]*PolyGamma[1, b+c] -
            2*PolyGamma[0, a+c]*PolyGamma[1, a+b+c+d] -
            2*PolyGamma[0, b+c]*PolyGamma[1, a+b+c+d] -
            PolyGamma[0, a+d]*PolyGamma[1, a+b+c+d] -
            PolyGamma[0, b+d]*PolyGamma[1, a+b+c+d] +
            3*PolyGamma[0, a+b+c+d]*PolyGamma[1, a+b+c+d] -
            PolyGamma[2, a+b+c+d]))/Gamma[a+b+c+d],
     Gamma[a + z]*Gamma[b + z]*Gamma[c - z]*Gamma[d - z] \
        PolyGamma[0,a+z] PolyGamma[0,b+z] PolyGamma[0,d-z] ->
         (Gamma[a+c]*Gamma[b+c]*Gamma[a+d]*Gamma[b+d]*
           (PolyGamma[0, a+c]*PolyGamma[0, b+c]*PolyGamma[0, a+d] +
            PolyGamma[0, b+c]*PolyGamma[0, a+d]^2+PolyGamma[0, a+c]*
             PolyGamma[0, b+c]*PolyGamma[0, b+d]+PolyGamma[0, a+c]*
             PolyGamma[0, a+d]*PolyGamma[0, b+d]+PolyGamma[0, b+c]*
             PolyGamma[0, a+d]*PolyGamma[0, b+d]+PolyGamma[0, a+d]^2*
             PolyGamma[0, b+d]+PolyGamma[0, a+c]*PolyGamma[0, b+d]^2 +
            PolyGamma[0, a+d]*PolyGamma[0, b+d]^2 - PolyGamma[0, a+c]*
             PolyGamma[0, b+c]*PolyGamma[0, a+b+c+d] -
            PolyGamma[0, a+c]*PolyGamma[0, a+d]*PolyGamma[0, a+b+c+d] -
            2*PolyGamma[0, b+c]*PolyGamma[0, a+d]*PolyGamma[0, a+b+c+d] -
            PolyGamma[0, a+d]^2*PolyGamma[0, a+b+c+d] -
            2*PolyGamma[0, a+c]*PolyGamma[0, b+d]*PolyGamma[0, a+b+c+d] -
            PolyGamma[0, b+c]*PolyGamma[0, b+d]*PolyGamma[0, a+b+c+d] -
            3*PolyGamma[0, a+d]*PolyGamma[0, b+d]*PolyGamma[0, a+b+c+d] -
            PolyGamma[0, b+d]^2*PolyGamma[0, a+b+c+d] +
            PolyGamma[0, a+c]*PolyGamma[0, a+b+c+d]^2 +
            PolyGamma[0, b+c]*PolyGamma[0, a+b+c+d]^2 +
            2*PolyGamma[0, a+d]*PolyGamma[0, a+b+c+d]^2 +
            2*PolyGamma[0, b+d]*PolyGamma[0, a+b+c+d]^2 -
            PolyGamma[0, a+b+c+d]^3+PolyGamma[0, b+c]*PolyGamma[1, a+d] +
            PolyGamma[0, b+d]*PolyGamma[1, a+d] - PolyGamma[0, a+b+c+d]*
             PolyGamma[1, a+d]+PolyGamma[0, a+c]*PolyGamma[1, b+d] +
            PolyGamma[0, a+d]*PolyGamma[1, b+d] - PolyGamma[0, a+b+c+d]*
             PolyGamma[1, b+d] - PolyGamma[0, a+c]*PolyGamma[1, a+b+c+d] -
            PolyGamma[0, b+c]*PolyGamma[1, a+b+c+d] -
            2*PolyGamma[0, a+d]*PolyGamma[1, a+b+c+d] -
            2*PolyGamma[0, b+d]*PolyGamma[1, a+b+c+d] +
            3*PolyGamma[0, a+b+c+d]*PolyGamma[1, a+b+c+d] -
            PolyGamma[2, a+b+c+d]))/Gamma[a+b+c+d],
     Gamma[a + z]*Gamma[b + z]*Gamma[c - z]*Gamma[d - z] \
        PolyGamma[0,a+z] PolyGamma[0,c-z] PolyGamma[0,d-z] ->
         (Gamma[a+c]*Gamma[b+c]*Gamma[a+d]*Gamma[b+d]*
           (PolyGamma[0, a+c]^2*PolyGamma[0, a+d]+PolyGamma[0, a+c]*
             PolyGamma[0, b+c]*PolyGamma[0, a+d]+PolyGamma[0, a+c]*
             PolyGamma[0, a+d]^2+PolyGamma[0, b+c]*PolyGamma[0, a+d]^2 +
            PolyGamma[0, a+c]^2*PolyGamma[0, b+d]+PolyGamma[0, a+c]*
             PolyGamma[0, b+c]*PolyGamma[0, b+d]+PolyGamma[0, a+c]*
             PolyGamma[0, a+d]*PolyGamma[0, b+d]+PolyGamma[0, b+c]*
             PolyGamma[0, a+d]*PolyGamma[0, b+d] - PolyGamma[0, a+c]^2*
             PolyGamma[0, a+b+c+d] - PolyGamma[0, a+c]*PolyGamma[0, b+c]*
             PolyGamma[0, a+b+c+d] - 3*PolyGamma[0, a+c]*PolyGamma[0, a+d]*
             PolyGamma[0, a+b+c+d] - 2*PolyGamma[0, b+c]*PolyGamma[0, a+d]*
             PolyGamma[0, a+b+c+d] - PolyGamma[0, a+d]^2*
             PolyGamma[0, a+b+c+d] - 2*PolyGamma[0, a+c]*PolyGamma[0, b+d]*
             PolyGamma[0, a+b+c+d] - PolyGamma[0, b+c]*PolyGamma[0, b+d]*
             PolyGamma[0, a+b+c+d] - PolyGamma[0, a+d]*PolyGamma[0, b+d]*
             PolyGamma[0, a+b+c+d]+2*PolyGamma[0, a+c]*
             PolyGamma[0, a+b+c+d]^2+PolyGamma[0, b+c]*
             PolyGamma[0, a+b+c+d]^2+2*PolyGamma[0, a+d]*
             PolyGamma[0, a+b+c+d]^2+PolyGamma[0, b+d]*
             PolyGamma[0, a+b+c+d]^2 - PolyGamma[0, a+b+c+d]^3 +
            PolyGamma[0, a+d]*PolyGamma[1, a+c]+PolyGamma[0, b+d]*
             PolyGamma[1, a+c] - PolyGamma[0, a+b+c+d]*PolyGamma[1, a+c] +
            PolyGamma[0, a+c]*PolyGamma[1, a+d]+PolyGamma[0, b+c]*
             PolyGamma[1, a+d] - PolyGamma[0, a+b+c+d]*PolyGamma[1, a+d] -
            2*PolyGamma[0, a+c]*PolyGamma[1, a+b+c+d] -
            PolyGamma[0, b+c]*PolyGamma[1, a+b+c+d] -
            2*PolyGamma[0, a+d]*PolyGamma[1, a+b+c+d] -
            PolyGamma[0, b+d]*PolyGamma[1, a+b+c+d] +
            3*PolyGamma[0, a+b+c+d]*PolyGamma[1, a+b+c+d] -
            PolyGamma[2, a+b+c+d]))/Gamma[a+b+c+d],
     Gamma[a + z]*Gamma[b + z]*Gamma[c - z]*Gamma[d - z] \
        PolyGamma[0,b+z] PolyGamma[0,c-z] PolyGamma[0,d-z] ->
         (Gamma[a+c]*Gamma[b+c]*Gamma[a+d]*Gamma[b+d]*
           (PolyGamma[0, a+c]*PolyGamma[0, b+c]*PolyGamma[0, a+d] +
            PolyGamma[0, b+c]^2*PolyGamma[0, a+d]+PolyGamma[0, a+c]*
             PolyGamma[0, b+c]*PolyGamma[0, b+d]+PolyGamma[0, b+c]^2*
             PolyGamma[0, b+d]+PolyGamma[0, a+c]*PolyGamma[0, a+d]*
             PolyGamma[0, b+d]+PolyGamma[0, b+c]*PolyGamma[0, a+d]*
             PolyGamma[0, b+d]+PolyGamma[0, a+c]*PolyGamma[0, b+d]^2 +
            PolyGamma[0, b+c]*PolyGamma[0, b+d]^2 - PolyGamma[0, a+c]*
             PolyGamma[0, b+c]*PolyGamma[0, a+b+c+d] -
            PolyGamma[0, b+c]^2*PolyGamma[0, a+b+c+d] -
            PolyGamma[0, a+c]*PolyGamma[0, a+d]*PolyGamma[0, a+b+c+d] -
            2*PolyGamma[0, b+c]*PolyGamma[0, a+d]*PolyGamma[0, a+b+c+d] -
            2*PolyGamma[0, a+c]*PolyGamma[0, b+d]*PolyGamma[0, a+b+c+d] -
            3*PolyGamma[0, b+c]*PolyGamma[0, b+d]*PolyGamma[0, a+b+c+d] -
            PolyGamma[0, a+d]*PolyGamma[0, b+d]*PolyGamma[0, a+b+c+d] -
            PolyGamma[0, b+d]^2*PolyGamma[0, a+b+c+d] +
            PolyGamma[0, a+c]*PolyGamma[0, a+b+c+d]^2 +
            2*PolyGamma[0, b+c]*PolyGamma[0, a+b+c+d]^2 +
            PolyGamma[0, a+d]*PolyGamma[0, a+b+c+d]^2 +
            2*PolyGamma[0, b+d]*PolyGamma[0, a+b+c+d]^2 -
            PolyGamma[0, a+b+c+d]^3+PolyGamma[0, a+d]*PolyGamma[1, b+c] +
            PolyGamma[0, b+d]*PolyGamma[1, b+c] - PolyGamma[0, a+b+c+d]*
             PolyGamma[1, b+c]+PolyGamma[0, a+c]*PolyGamma[1, b+d] +
            PolyGamma[0, b+c]*PolyGamma[1, b+d] - PolyGamma[0, a+b+c+d]*
             PolyGamma[1, b+d] - PolyGamma[0, a+c]*PolyGamma[1, a+b+c+d] -
            2*PolyGamma[0, b+c]*PolyGamma[1, a+b+c+d] -
            PolyGamma[0, a+d]*PolyGamma[1, a+b+c+d] -
            2*PolyGamma[0, b+d]*PolyGamma[1, a+b+c+d] +
            3*PolyGamma[0, a+b+c+d]*PolyGamma[1, a+b+c+d] -
            PolyGamma[2, a+b+c+d]))/Gamma[a+b+c+d]};

  (* Include "xx" to be ready for shifting parameters *)
  shift = {a->a+xx,b->b+xx,c->c+xx,d->d+xx};
  patterns = #[[1]]->(#[[2]] /. shift)& /@ patterns;
  (* For the analytic continuation *)
  continuations = #[[1]]->(#[[1]] /. shift)& /@ patterns;

  integ = integ /. f_[x_] :> f[Expand[x]] /; MemberQ[{Gamma,PolyGamma},f];
  pos = Cases[{Numerator[integrand]},
    Gamma[x_] :> x /; Coefficient[x, z] == +1, -1] /. z -> 0;
  If[Length[pos] == 1, pos = {First[pos], First[pos]}];
  If[Length[pos] != 2, Return[original]];
  a = First[pos];
  b = Last[pos];

  neg = Cases[{Numerator[integrand]},
    Gamma[x_] :> x /; Coefficient[x, z] == -1, -1] /. z -> 0;
  If[Length[neg] == 1, neg = {First[neg], First[neg]}];
  If[Length[neg] != 2, Return[original]];
  c = First[neg];
  d = Last[neg];

  patterns = patterns /. 
                f_[x_] :> f[Expand[x]] /; MemberQ[{Gamma,PolyGamma},f];
  (* Does any pattern match all appearances of "z"? *)
  pattern = Scan[If [!MemberQ[integ /. #,z,-1],Return[#]]&,
                 patterns];
  If [verbosity >= 3,Print[pattern]];

  (* Nope... just return original expression *)
  If [pattern === Null,
     If [recordFailedBarnes,failedBarnes1 = Join[failedBarnes1,{original}]];
     Return[original]];

  (* Determine minimum xx value which will make all Gamma arguments positive,
     hence the integral well-defined on the given contour with properly
     separated poles *)
  p = -Min[pos /. rules];
  n = +Min[neg /. rules];
  xx0 = Max[{p - z, z - n} /. rules];

  (* No poles crossed going to xx = 0: *)
  If[xx0 < 0,
    Return[DeleteCases[original /. (pattern /. xx->0),
      z -> _, -1]]];

  (* Otherwise, we Need to shift "a", etc; take matching continuation 
     to our resolution *)
  continue = continuations[[Flatten[Position[patterns,pattern]]]];
  int = integ /. continue;

  fixed = DeleteCases[Join[fixedVarRules, intVarRules, {xx -> xx0 + 10^-10}],
    z -> _, -1];

  cont = MBmerge[MBcontinue[int, xx -> 0, {fixed, {z -> (z /. intVarRules)}},
    Verbose -> False]];

  ct = -cont[[1,1]];
  orig = integ /. pattern;

  If [!FreeQ[orig,z] || !FreeQ[ct,z],Print["Barnes1 oops!"];
Print[InputForm[orig]];
Print[InputForm[cont]];
Print[InputForm[ct]];
Print[integ];
Print[pattern];
      Throw["oops!"]];

  (* Cannot handle cases with poles in xx;  for obscure reasons,
     Series may hang on orig+ct (even when it succeeds on each term
     separately); so first do the Series on each term separately; even
     just adding the two series can cause it to hang, so be careful there
     too) *)
  (*
  int = Normal[Simplify[Series[Normal[Series[orig,{xx,0,0}]] 
                               + Normal[Series[ct,{xx,0,0}]], {xx, 0, 0}]]];
  *)
  int = (Normal[Series[orig,{xx,0,0}]] + Normal[Series[ct,{xx,0,0}]]);
  If [verbosity >= 3,Print["Done 1st stage series"]];
  (* It can happen that the coefficient of 1/xx is zero but not manifestly
     so, because of different PolyGammas appearing *)
  int = int /. xx^p_. x_ :> xx^p FactorSquareFree[CanonicalGamma[x,
                               Complement[#[[1]]& /@ intVarRules,{z}]]];
  int = Normal[Simplify[Series[int, {xx, 0, 0}]]];

  If [!FreeQ[int,xx],
     If [recordFailedBarnes,failedBarnes1 = Join[failedBarnes1,{original}]];
     Return[original]];

  Return[MBint[int,
    {fixedVarRules, DeleteCases[intVarRules, z -> _, -1]}]]]
 
Barnes2[MBint[integrand_, {fixedVarRules_List, intVarRules_List}], z_Symbol] := 
Block[{original = MBint[integrand, {fixedVarRules, intVarRules}], sign = +1,
  integ = integrand, rules = Flatten[{fixedVarRules, intVarRules}], x, i, j,
  tpos, pos = {}, a, b, c, p, tneg, neg = {}, d, e, n, 
  xx, xx0, int, fixed, cont, ct, orig,
  patterns, pattern, continuations, continue, patterns2, continuations2,
  shift, shift2, constraint, cvars, fixCoeff, result},

  (* Modified 9/25/06 (DAK) to match forms with f=a+b+c+d+e+small int *)
  (* "patterns" is a list of pattern->replacement entries; note that
     the values of "a", "b", "c", "d", "e", "f" are determined
     below. 
   *)
  patterns = {
      Gamma[a + z]*Gamma[b + z]*Gamma[c + z]*Gamma[d - z]*Gamma[e - z]/
      Gamma[Expand[a + b + c + d  + e] + z] ->
      Gamma[a + d]*Gamma[b + d]*Gamma[c + d]*
      Gamma[a + e]*Gamma[b + e]*Gamma[c + e]/
      (Gamma[a + b + d + e]*Gamma[a + c + d + e]*Gamma[b + c + d + e]),
      (* From Appendix D of Smirnov's book, or his barnes.m *)
      Gamma[a + z]*Gamma[b + z]*Gamma[c + z]*Gamma[d - z]*Gamma[e - z]/
      Gamma[Expand[a + b + c + d  + e]+1 + z] ->
      (Gamma[b + d]*Gamma[a + e]*
    (-(Gamma[1 + a + d]*Gamma[1 + c + d]*Gamma[-a - c - d - e]*Gamma[b + e]) +
      Gamma[a + d]*Gamma[c + d]*Gamma[1 - a - c - d - e]*Gamma[1 + b + e])*
      Gamma[c + e])/(Gamma[1 - a - c - d - e]*Gamma[1 + a + b + d + e]*
       Gamma[a + c + d + e]*Gamma[1 + b + c + d + e]),
      (* From Appendix D of Smirnov's book, or his barnes.m *)
      Gamma[a + z]*Gamma[b + z]*Gamma[c + z]*Gamma[d - z]*Gamma[e - z]/
      Gamma[Expand[a + b + c + d  + e]+2 + z] ->
         (Gamma[b + d]*Gamma[a + e]*(Gamma[2 + a + d]*Gamma[2 + c + d]*
         Gamma[-1 - a - c - d - e]*Gamma[b + e] - 2*Gamma[1 + a + d]*
         Gamma[1 + c + d]*Gamma[-a - c - d - e]*Gamma[1 + b + e] +
       Gamma[a + d]*Gamma[c + d]*Gamma[1 - a - c - d - e]*Gamma[2 + b + e])*
      Gamma[c + e])/(Gamma[1 - a - c - d - e]*Gamma[2 + a + b + d + e]*
       Gamma[a + c + d + e]*Gamma[2 + b + c + d + e])};
  patterns2 = {
      (* From Smirnov's Mel51ozIR, residual = -1, modified to make a<->b
         symmetry manifest *)
      Gamma[a+z] Gamma[b+z] Gamma[c+z] Gamma[Expand[2-a-b-e]-z] Gamma[e-z]/
             Gamma[1+c+z] -> 
       -Gamma[1 - a - e]*Gamma[1 - b - e]*Gamma[a + e]*Gamma[b + e] *
        (1-Gamma[2-a-b-e + c]*Gamma[e + c]/(Gamma[1-a + c]*Gamma[1-b + c])),
      (* Similar with b<->c *)
      Gamma[a+z] Gamma[c+z] Gamma[b+z] Gamma[Expand[2-a-c-e]-z] Gamma[e-z]/
             Gamma[1+b+z] -> 
       -Gamma[1 - a - e]*Gamma[1 - c - e]*Gamma[a + e]*Gamma[c + e] *
        (1-Gamma[2-a-c-e + b]*Gamma[e + b]/(Gamma[1-a + b]*Gamma[1-c + b])),
       (* And a<->c *)
      Gamma[c+z] Gamma[b+z] Gamma[a+z] Gamma[Expand[2-c-b-e]-z] Gamma[e-z]/
             Gamma[1+a+z] -> 
       -Gamma[1 - b - e]*Gamma[1 - c - e]*Gamma[b + e]*Gamma[c + e] *
        (1-Gamma[2-b-c-e+a]*Gamma[e+a]/(Gamma[1-b+a]*Gamma[1-c+a]))
       };

  (* Include "xx" to be ready for shifting parameters *)
  shift = {a->a+xx,b->b+xx,c->c+xx,d->d+xx,e->e+xx};
  patterns = #[[1]]->(#[[2]] /. shift)& /@ patterns;
  (* For the analytic continuation *)
  continuations = #[[1]]->(#[[1]] /. shift)& /@ patterns;

  (* In patterns2, we need different analytic continuations, because the
     Gamma functions involve e-c as well as a+c etc; we leave the
     coefficients open in order to maximize the chance of finding
     a region in which we can define the integral... *)
  shift2 = {a->a+ca xx,b->b+cb xx,d->d-(ca+cb-ce) xx,c->c+cc xx,e->e-ce xx};
  patterns2 = #[[1]]->(#[[2]] /. shift2)& /@ patterns2;
  continuations2 = #[[1]]->(#[[1]] /. shift2)& /@ patterns2;

  patterns = Join[patterns,patterns2];
  continuations = Join[continuations,continuations2];

  integ = integ /. f_[x_] :> f[Expand[x]] /; MemberQ[{Gamma,PolyGamma},f];
  If[MemberQ[{Denominator[integrand]}, Gamma[x_.-z], -1],
    sign = -1;
    integ = integ /. z -> -z;
    rules = rules /. (z -> x_) -> (z -> -x)];

  tpos = Cases[{Numerator[integ]},
    Gamma[x_] :> x /; Coefficient[x, z] == +1, -1] /. z -> 0;

  Do[
    Do[AppendTo[pos, tpos[[i]]],
    {j, Exponent[integ, Gamma[tpos[[i]] + z]]}],
  {i, Length[tpos]}];

  If[Length[pos] != 3, Return[original]]; 

  a = pos[[1]];
  b = pos[[2]];
  c = pos[[3]];

  tneg = Cases[{Numerator[integ]},
    Gamma[x_] :> x /; Coefficient[x, z] == -1, -1] /. z -> 0;

  Do[
    Do[AppendTo[neg, tneg[[i]]],
    {j, Exponent[integ, Gamma[tneg[[i]] - z]]}],
  {i, Length[tneg]}];

  If[Length[neg] != 2, Return[original]]; 

  d = neg[[1]];
  e = neg[[2]];

  patterns = patterns /. 
                f_[x_] :> f[Expand[x]] /; MemberQ[{Gamma,PolyGamma},f];
  patterns2 = patterns2 /. 
                f_[x_] :> f[Expand[x]] /; MemberQ[{Gamma,PolyGamma},f];

  (* Does any pattern match all appearances of "z"? *)
  pattern = Scan[If [!MemberQ[integ /. #,z,-1],Return[#]]&,
                 patterns];

  extraCare = MemberQ[patterns2,pattern];
  If [verbosity >= 3,
      Print["Picked pattern ",Flatten[Position[patterns,pattern]][[1]],
            " of ",Length[patterns]]];

  (* Nope... just return original expression *)
  If [pattern === Null,
     If [recordFailedBarnes,failedBarnes2 = Join[failedBarnes2,{original}]];
     Return[original]];

  (* Find the minimum xx value which makes all Gammas positive. 
     If it's negative, no poles are crossed on our way to xx = 0. *)
  (* The original code assumed the coefficient of xx in all parameters
     was +1:
  p = -Min[pos /. rules];
  n = +Min[neg /. rules];
  xx0 = Max[{p - z, z - n} /. rules];
   
  If[xx0 < 0,
    Return[DeleteCases[original /. z -> sign*z /. (pattern /. xx->0),
      sign*z -> _, -1]]];

    but we need to take a more general approach, because in the last
    set of patterns, the coefficient is different from 1.
   *)
  continue = continuations[[Flatten[Position[patterns,pattern]]]];
  constraint = 
    Select[Variables[Numerator[integ /. continue]],!FreeQ[#,xx]&] /. rules;
  cvars = Union[Cases[constraint,_Symbol,-1]];

  constraint = (And @@ (Last[#]>0 & /@ constraint));

  (* It's expensive to reduce this in generality.  Does a simple choice
     for ca, cb, cc, ce work? *)
  fixCoeff = {ca->1,cb->1,cc->1,ce->1};
  If[TrueQ[!(reduced = Reduce[constraint /. fixCoeff,xx])],
     (* False for all xx; need to work harder. *)
     (* Can we find a solution with xx->0? *)
     If [constraint /. xx->0,constraint = True,
     (* Otherwise, we hunt... and would like to pick the one with 
        smallest absolute value for xx.  However, this appears to 
        ultimately cause a memory fault [core dump] (!) in Mathematica 5.2 
        when taking residues, if the size of the numerator and denominator 
        in the rational fraction are too large; so try to minimize them 
        instead *)
     (* It's better to use strict rather than weak inequalities... (< vs <=) *)
     bound = And @@ ((# != 0 && -1 < # < 1)& /@ Select[cvars,#=!=xx&]);
     If [verbosity >= 3,Print["Looking for instance"]];
     (* Try to find a lone solution first; it's much faster *)
     fixCoeff = FindInstance[constraint && bound,cvars,Rationals];
     If [fixCoeff != {},
     TimeConstrained[
     fixCoeff = \
       Sort[FindInstance[constraint && bound,cvars,Rationals,10],
            Abs[Denominator[ ca cb cc ce xx /. #1]] < 
                Abs[Denominator[ca cb cc ce xx /. #2]]&],searchTimeLimit,{}]];
     If [verbosity >= 3,Print["Found ",Length[fixCoeff]," instance(s)"]];
     If [fixCoeff == {}, (* No possible solution *) constraint = False,
        (* Just fix the parameters ca, cb, cc, ce for now *)
        fixCoeff = DeleteCases[fixCoeff[[1]],xx->_,-1];
        constraint = constraint /. fixCoeff;
        constraint = Reduce[constraint,xx]]],
      (* otherwise, simple choice is OK *) constraint=reduced];
  
  (* If it can't be satisfied, punt *)
  If [TrueQ[!constraint],
     Print["Warning: no separation of poles possible in Barnes2"];
     If [recordFailedBarnes,failedBarnes2 = Join[failedBarnes2,{original}]];
     Return[original]];

  If [constraint /. xx->0 (* No poles between proper definition & origin *),
    (* We need to be careful, because just substituting xx->0 might
       (very occasionally) lead to an ill-defined result; this seems to
       happen with the last three patterns; so we do a series expansion *)
    result = DeleteCases[original /. z -> sign*z /. pattern,
      sign*z -> _, -1];
    If [extraCare,
      result = MBint[
           FactorSquareFree[Normal[Series[result[[1]],{xx,0,0}]]],
           result[[2]]],
      result = result /. xx->0];
    If [!FreeQ[result,xx],
       If [recordFailedBarnes,failedBarnes2 = Join[failedBarnes2,{original}]];
       Return[original]];
    Return[result]];

  (* Otherwise, need to shift "a", etc; use matching continuation 
     to our pattern.  Pick the "simplest" xx0 within the range *)
  xx0 = Sort[Flatten[xx /. FindInstance[constraint,cvars,20]],
             Abs[Denominator[#1]]<Abs[Denominator[#2]]&][[1]];
  If [verbosity >= 3, Print["Picked xx0 = ",xx0]];
  int = integ /. continue;
  int = int /. fixCoeff;

  fixed = DeleteCases[Join[fixedVarRules, intVarRules, {xx -> xx0 + 10^-10}],
    z -> _, -1];

  (* "Residue" inside MBresidue inside MBcontinue can be a source of
     memory faults in Mathematica 5.2... watch out! *)
  cont = MBmerge[MBcontinue[int, xx -> 0,
    {fixed, {z -> (sign*z /. intVarRules)}}, Verbose -> verbosity >= 3]];

  ct = -cont[[1,1]];
  orig = integ /. pattern;
  If [extraCare,  (* See earlier discussion *)
     orig = FactorSquareFree /@ Series[orig,{xx,0,0}];
     ];  
  orig = Normal[orig];
  orig = orig /. fixCoeff;

  If [verbosity >= 3,Print["Obtained counterterm"]];
  If [!FreeQ[orig,z] || !FreeQ[ct,z],Print["Barnes2 oops!"];Throw["oops!"]];
  (* Cannot handle cases with poles in xx;  also, for obscure reasons,
     Series may hang on orig+ct (even when it succeeds on each term
     separately; so first do the Series on each term separately...) 
     10/7/06: That isn't always enough, so clean up in between... *)
  int = (Normal[Series[orig,{xx,0,0}]] + Normal[Series[ct,{xx,0,0}]]);
  If [verbosity >= 3,Print["Done 1st stage series"]];
  (* It can happen that the coefficient of 1/xx is zero but not manifestly
     so, because of different PolyGammas appearing *)
  int = int /. xx^p_. x_ :> xx^p FactorSquareFree[CanonicalGamma[x,
                               Complement[#[[1]]& /@ intVarRules,{z}]]];
  int = Normal[Simplify[Series[int, {xx, 0, 0}]]];

(*Print[xint=Normal[Series[int,{xx,0,-1}]]];*)
  If [!FreeQ[int,xx],
     If [recordFailedBarnes,failedBarnes2 = Join[failedBarnes2,{original}]];
     Return[original]];

  Return[MBint[int,
    {fixedVarRules, DeleteCases[intVarRules, z -> _, -1]}]]]

End[];
EndPackage[];
